import functools

from .circuit import _definition_context_stack
from .display import Display


class Log(Display):
    level = None

    def __init__(self, display_str, args, file=None):
        display_prefix = type(self).__name__.upper()
        super().__init__(f"[{display_prefix}] {display_str}", args, file=file)

    def _make_cond_str(self, format_args):
        cond_str = super()._make_cond_str(format_args)
        level_check = f"`MAGMA_LOG_LEVEL <= {self.level}"
        if not cond_str:
            return level_check
        return f"({level_check}) && ({cond_str})"


class Debug(Log):
    level = 0


class Info(Log):
    level = 1


class Warning(Log):
    level = 2


class Error(Log):
    level = 3


def _make_log_func(T):
    @functools.wraps(_make_log_func)
    def log_func(log_str, *args, file=None):
        context = _definition_context_stack.peek()
        log = T(log_str, args, file=file)
        context.add_display(log)
        context.insert_default_log_level()
        return log
    return log_func


debug = _make_log_func(Debug)
info = _make_log_func(Info)
warning = _make_log_func(Warning)
error = _make_log_func(Error)



import os
from magma.array import Array
from magma.config import get_compile_dir, set_compile_dir
from magma.digital import Digital
from magma.passes.passes import CircuitPass
from magma.tuple import Tuple
from magma.verilog_utils import value_to_verilog_name


def _gen_bind_port(cls, mon_arg, bind_arg):
    if isinstance(mon_arg, Tuple) or isinstance(mon_arg, Array) and \
            not issubclass(mon_arg.T, Digital):
        result = []
        for child1, child2 in zip(mon_arg, bind_arg):
            result += _gen_bind_port(cls, child1, child2)
        return result
    port = value_to_verilog_name(mon_arg)
    arg = value_to_verilog_name(bind_arg)
    return [(f".{port}({arg})")]


def _bind(cls, monitor, compile_fn, *args):
    bind_str = monitor.verilogFile
    ports = []
    for mon_arg, cls_arg in zip(monitor.interface.ports.values(),
                                cls.interface.ports.values()):
        if str(mon_arg.name) != str(cls_arg.name):
            error_str = f"""
Bind monitor interface does not match circuit interface
    Monitor Ports: {list(monitor.interface.ports)}
    Circuit Ports: {list(cls.interface.ports)}
"""
            raise TypeError(error_str)
        ports += _gen_bind_port(cls, mon_arg, cls_arg)
    extra_mon_args = list(
        monitor.interface.ports.values()
    )[len(cls.interface):]
    for mon_arg, bind_arg in zip(extra_mon_args, args):
        ports += _gen_bind_port(cls, mon_arg, bind_arg)
    ports_str = ",\n    ".join(ports)
    bind_str = f"bind {cls.name} {monitor.name} {monitor.name}_inst (\n    {ports_str}\n);"  # noqa
    if not os.path.isdir(".magma"):
        os.mkdir(".magma")
    curr_compile_dir = get_compile_dir()
    set_compile_dir("normal")
    # Circular dependency, need coreir backend to compile, backend imports
    # circuit (for wrap casts logic, we might be able to factor that out).
    compile_fn(f".magma/{monitor.name}", monitor, inline=True)
    set_compile_dir(curr_compile_dir)
    with open(f".magma/{monitor.name}.v", "r") as f:
        content = "\n".join((f.read(), bind_str))
    cls.bind_modules[monitor.name] = content


class BindPass(CircuitPass):
    def __init__(self, main, compile_fn):
        super().__init__(main)
        self._compile_fn = compile_fn

    def __call__(self, cls):
        if cls.bind_modules_bound:
            return
        bind_modules = cls.bind_modules.copy()
        cls.bind_modules = {}
        for monitor, args in bind_modules.items():
            _bind(cls, monitor, self._compile_fn, *args)
        cls.bind_modules_bound = True



from collections import OrderedDict
import json
from hwtypes import BitVector
from ..array import Array
from ..bit import Digital
from ..clock import Clock, AsyncReset, AsyncResetN
from ..ref import ArrayRef, DefnRef, TupleRef, InstRef
from ..tuple import Tuple
from ..protocol_type import MagmaProtocol
from .util import make_relative


class CoreIRBackendError(RuntimeError):
    pass


def _tuple_key_to_string(k):
    try:
        int(k)
        return f"_{k}"
    except ValueError:
        pass
    return k


def magma_name_to_coreir_select(name):
    if isinstance(name, InstRef):
        return f"{name.inst.name}.{name.name}"
    if isinstance(name, DefnRef):
        return f"self.{name.name}"
    if isinstance(name, ArrayRef):
        array_name = magma_name_to_coreir_select(name.array.name)
        return f"{array_name}.{name.index}"
    if isinstance(name, TupleRef):
        tuple_name = magma_name_to_coreir_select(name.tuple.name)
        key_name = _tuple_key_to_string(name.index)
        return f"{tuple_name}.{key_name}"
    raise NotImplementedError(name)


def magma_port_to_coreir_port(port):
    return magma_name_to_coreir_select(port.name)


def check_magma_type(type_, error_msg=""):
    if issubclass(type_, MagmaProtocol):
        type_ = type_._to_magma_()
    if issubclass(type_, Array):
        msg = error_msg.format("Array({}, {})").format(str(type_.N), "{}")
        check_magma_type(type_.T, msg)
        return
    if issubclass(type_, Tuple):
        for (k, t) in zip(type_.keys(), type_.types()):
            msg = error_msg.format("Tuple({}:{})".format(k, "{}"))
            check_magma_type(t, msg)
        return
    if issubclass(type_, Digital):
        return
    raise CoreIRBackendError(error_msg.format(str(type_)))


def check_magma_interface(interface):
    # For now only allow Bit, Array, or Tuple.
    for name, port in interface.ports.items():
        check_magma_type(
            type(port), f"{name}: {type(port)} not supported by CoreIR backend"
        )


def magma_type_to_coreir_type(context, type_):
    if issubclass(type_, MagmaProtocol):
        type_ = type_._to_magma_()

    if issubclass(type_, Array):
        return context.Array(type_.N,
                             magma_type_to_coreir_type(context, type_.T))
    if issubclass(type_, Tuple):
        zipped = zip(type_.keys(), type_.types())
        return context.Record({
            _tuple_key_to_string(k): magma_type_to_coreir_type(context, t)
            for (k, t) in zipped
        })
    if type_.is_input():
        if issubclass(type_, Clock):
            return context.named_types[("coreir", "clk")]
        if issubclass(type_, (AsyncReset, AsyncResetN)):
            return context.named_types[("coreir", "arst")]
        return context.Bit()
    if type_.is_output():
        if issubclass(type_, Clock):
            return context.named_types[("coreir", "clkIn")]
        if issubclass(type_, (AsyncReset, AsyncResetN)):
            return context.named_types[("coreir", "arstIn")]
        return context.BitIn()
    return context.BitInOut()


def magma_interface_to_coreir_module_type(context, interface):
    args = OrderedDict()
    for name, port in interface.ports.items():
        args[name] = magma_type_to_coreir_type(context, type(port))
    return context.Record(args)


def add_non_input_ports(non_input_ports, port):
    if isinstance(port, MagmaProtocol):
        port = port._get_magma_value_()
    if not port.is_input():
        non_input_ports[port] = magma_port_to_coreir_port(port)
    if isinstance(port, (Tuple, Array)):
        for element in port:
            add_non_input_ports(non_input_ports, element)


def python_to_coreir_param_type(context, typ):
    if typ is int:
        return context.Int()
    if typ is bool:
        return context.Bool()
    if typ is str:
        return context.String()
    if typ is BitVector:
        return context.BitVector()
    raise NotImplementedError(typ)


def make_cparams(context, params):
    cparams = {
        k: python_to_coreir_param_type(context, t)
        for k, t in params.items()
    }
    return context.newParams(cparams)


def is_clock_or_nested_clock(p):
    if issubclass(p, Clock):
        return True
    if issubclass(p, Array):
        return is_clock_or_nested_clock(p.T)
    if issubclass(p, Tuple):
        for item in p.types():
            if is_clock_or_nested_clock(item):
                return True
    return False


def attach_debug_info(coreir_obj, debug_info, a=None, b=None):
    def fn(k, v):
        if a and b:
            return coreir_obj.add_metadata(a, b, k, v)
        return coreir_obj.add_metadata(k, v)

    fn("filename", json.dumps(make_relative(debug_info.filename)))
    fn("lineno", json.dumps(str(debug_info.lineno)))


def map_genarg(context, value):
    if isinstance(value, type) and issubclass(value, Clock):
        if value.is_input():
            return context.named_types[("coreir", "clkIn")]
        return context.named_types[("coreir", "clk")]
    if isinstance(value, type) and issubclass(value, (AsyncReset, AsyncResetN)):
        if value.is_input():
            return context.named_types[("coreir", "arstIn")]
        return context.named_types[("coreir", "arst")]
    return value


def get_module_of_inst(context, inst, lib):
    wrapped = getattr(inst, "wrappedModule", None)
    if wrapped and wrapped.context is context:
        return wrapped
    return lib.modules[type(inst).coreir_name]


def get_inst_args(inst):
    args = {}
    for name, value in inst.kwargs.items():
        if name == "name" or name == "loc":
            continue
        if isinstance(value, tuple):
            args[name] = BitVector[value[1]](value[0])
        else:
            args[name] = value
    return args


def constant_to_value(constant):
    assert constant.const()
    if isinstance(constant, Digital):
        return 1 if constant is type(constant).VCC else 0
    if isinstance(constant, Array):
        values = [constant_to_value(c) for c in constant]
        return BitVector[len(constant)](values)
    raise NotImplementedError(constant)



from .clock import AsyncReset, AsyncResetN, Clock, Enable, Reset
from .interface import IO
from .t import In


def ClockIO(has_enable=False, has_reset=False, has_ce=False,
            has_async_reset=False, has_async_resetn=False):
    args = {'CLK': In(Clock)}
    has_enable |= has_ce
    if has_enable:
        args['CE'] = In(Enable)
    if has_reset:
        args['RESET'] = In(Reset)
    if has_async_reset:
        args['ASYNCRESET'] = In(AsyncReset)
    if has_async_resetn:
        args['ASYNCRESETN'] = In(AsyncResetN)
    return IO(**args)



import os
import subprocess
from .coreir_ import InsertWrapCasts
from ..compiler import Compiler
from ..frontend import coreir_ as coreir_frontend
from ..passes import InstanceGraphPass
from ..passes.insert_coreir_wires import InsertCoreIRWires
from ..logging import root_logger


_logger = root_logger()


def _make_verilog_cmd(deps, basename, opts):
    cmd = "coreir"
    if deps:
        cmd += f" -l {','.join(deps)}"
    cmd += f" -i {basename}.json"
    if opts.get("split", ""):
        split = opts["split"]
        cmd += f" -o \"{split}/*.v\" -s"
    else:
        cmd += f" -o {basename}.v"
    if opts.get("inline", False):
        cmd += " --inline"
    if opts.get("verilator_debug", False):
        cmd += " --verilator_debug"
    return cmd


class CoreIRCompiler(Compiler):
    def __init__(self, main, basename, opts):
        super().__init__(main, basename, opts)
        self.namespaces = self.opts.get("namespaces", ["global"])
        self.passes = opts.get("passes", [])
        if "markdirty" not in self.passes:
            self.passes.append("markdirty")
        self.deps = self.__deps()

    def compile(self):
        self.__compile_coreir()
        if self.opts.get("output_verilog", False):
            self.__compile_verilog()
            return
        if self.opts.get("header_file", False):
            _logger.warning("[coreir-compiler] header_file only supported "
                            "when output_verilog=True, ignoring")
        if self.opts.get("header_str", False):
            _logger.warning("[coreir-compiler] header_str only supported "
                            "when output_verilog=True, ignoring")
        if self.opts.get("footer_str", False):
            _logger.warning("[coreir-compiler] footer_str only "
                            "supported when output_verilog=True, ignoring")

    def __compile_coreir(self):
        backend = coreir_frontend.GetCoreIRBackend()
        InsertWrapCasts(self.main).run()
        InsertCoreIRWires(self.main).run()
        backend.compile(self.main)
        backend.context.run_passes(self.passes, self.namespaces)
        main_key = self.main.coreir_name
        backend.modules[main_key].save_to_file(self.basename + ".json")

    def __compile_verilog(self):
        # TODO(rsetaluri): Expose compilation to verilog in pycoreir rather than
        # calling the binary from the command line.
        cmd = _make_verilog_cmd(self.deps, self.basename, self.opts)

        ret = subprocess.run(cmd, shell=True).returncode
        if ret:
            raise RuntimeError(f"CoreIR cmd '{cmd}' failed with code {ret}")
        self._process_header_footer()

        backend = coreir_frontend.GetCoreIRBackend()
        # TODO: We need fresh bind_files for each compile call
        for name, file in backend.sv_bind_files.items():
            filename = os.path.join(os.path.dirname(self.basename), name)
            with open(f"{filename}.sv", "w") as f:
                f.write(file)

        if self.opts.get("sv", False):
            subprocess.run(["mv", f"{self.basename}.v", f"{self.basename}.sv"])

    def __deps(self):
        deps = self.opts.get("coreir_libs", set())
        pass_ = InstanceGraphPass(self.main)
        pass_.run()
        for key, _ in pass_.tsortedgraph:
            if key.coreir_lib:
                deps.add(key.coreir_lib)
            elif hasattr(key, "wrappedModule"):
                deps |= key.coreir_wrapped_modules_libs_used
        deps |= set(self.namespaces)
        return deps

    def _process_header_footer(self):
        header = ""
        footer = ""
        if self.opts.get("header_file", False):
            with open(self.opts["header_file"], "r") as header_file:
                header = header_file.read() + "\n"
        if self.opts.get("header_str", False):
            header += self.opts["header_str"] + "\n"
        if self.opts.get("footer_str", False):
            footer = self.opts["footer_str"]
        if header is not "" or footer is not "":
            if self.opts.get("split", False):
                _logger.warning(
                    "header/footer logic not supported with split, ignoring")
                return
            with open(f"{self.basename}.v", "r") as verilog_file:
                verilog = verilog_file.read()
            with open(f"{self.basename}.v", "w") as verilog_file:
                verilog_file.write(header + verilog + "\n" + footer)



import types
from collections import OrderedDict
from collections.abc import Sequence
from ..bitutils import seq2int, fun2seq
from ..bit import Bit, Digital
from ..array import Array
from ..compiler import make_compiler
from ..clock import wiredefaultclock
from ..is_definition import isdefinition

__all__ = ['compile']

def fullname(prefix, name):
    return prefix + '_' + name if prefix else name

def fullqualifiedname(prefix, v):
     return fullname(prefix, str(v))

def fullinputname(prefix, v):
    if isinstance(v, Digital) and v.const():
        return "$true" if v is type(v).VCC else "$false"
    return fullqualifiedname(prefix, v)

def compileinput(prefix, from_, to):
    assert isinstance(from_, Bit)
    to = fullqualifiedname(prefix, to)
    from_ = fullinputname(prefix, from_)
    return '.names %s %s\n1 1\n' % (from_, to)

def compileinputs(instance, prefix):
    s = ''
    for k, v in instance.interface.ports.items():
        if v.is_input():
            if isinstance(v, Array):
                for i in range(len(v)):
                    w = v[i].value()
                    if w:
                        s += compileinput(prefix, w, v[i])
            else:
                w = v.value()
                if w:
                    s += compileinput(prefix, w, v)
    return s


def compileprimitive(prim, prefix, location):

    assert not isdefinition(prim)

    def arg(k,v): # helper function to format arg
        if not isinstance(v, str): v = str(v)
        return '%s=%s' % (k, v)

    # collect args
    args = []
    for k, v in prim.interface.ports.items():
        if v.is_input() or v.is_inout():
            # connect input to an output by tracing the net
            w = v.value()
            if w:
                w = fullinputname(prefix, w)
                args.append( arg(k,w) )
        elif v.is_output() or v.is_inout():
            # connect output to a wire with a fully qualified name
            w = fullqualifiedname(prefix, v)
            args.append( arg(k,w) )

    s  = '.gate ' + str(prim.__class__.__name__) + ' ' + ' '.join(args) + '\n'

    # parameters
    for k, v in prim.kwargs.items():
        if k == 'LUT_INIT':
            if isinstance(v, tuple):
                v = v[0]
            #v = int('0x'+v[4:], 0) # convert from verilog hex
            v = bin(v)[2:]         # convert to binary, skipping 0b
            s += '.param %s %s\n' % (k, v)

    # attributes - only location for now
    if location:
        s += '.attr loc "%d,%d/%d"\n' % location

    return s


def compiledefinition(defn, prefix, location=None):
    s = ''
    for instance in defn.instances:
        wiredefaultclock(defn, instance)
        if location and instance.loc:
            x = location[0] + instance.loc[0]
            y = location[1] + instance.loc[1]
            z = location[2] + instance.loc[2]
            instlocation = (x,y,z)
        else:
            instlocation = None
        s += compileinstance(instance, prefix, instlocation)
    s += '# wire instance outputs\n'
    return s + compileinputs(defn, prefix)

def compileinstance(instance, prefix, location):
    defn = type(instance)
    if isdefinition(defn):
        instname = fullname(prefix, instance.name)
        s  = '# ' + str(defn) + ' ' + instname + ' instances\n'
        s += compiledefinition(defn, instname, location)
        s += '# wire instance inputs\n'
        s += compileinputs(instance, prefix)
        s += '# ' + str(defn) + ' ' + instname + '\n'
    else:
        s  = compileprimitive(instance, prefix, location)
    return s

def compile(main, origin=None):

    s = '.model %s\n' % main.__name__

    inputs = []
    outputs = []
    for k, v in main.interface.ports.items():
        if v.is_input() or v.s_inout():
            inputs.append(k)
        if v.is_output() or v.is_inout():
            outputs.append(k)
    s += '.inputs '  + " ".join(outputs) + '\n' # flip inputs
    s += '.outputs ' + " ".join(inputs) + '\n'  # flip outputs
         
    s += '.names $false\n'
    s += '.names $true\n'
    s += '1\n'

    if origin and len(origin) == 2:
        origin = (origin[0], origin[1], 0)
    s += compiledefinition(main, '', origin)

    return s


BlifCompiler = make_compiler("blif", compile)



try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache

cachedFunctions = []


def clear_cachedFunctions():
    for func in cachedFunctions:
        func.cache_clear()


def cache_definition(fn):
    cachedFunctions.append(lru_cache(maxsize=None)(fn))
    return cachedFunctions[-1]

def singleton(cls):
    instance = [None]
    def wrapper(*args, **kwargs):
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        return instance[0]

    return wrapper

from magma.protocol_type import MagmaProtocolMeta, MagmaProtocol

# wires
from .wire import wire

# types
from .t import *

from .bit import *

from .array import *
from .bits import *
from .bfloat import BFloat
from .tuple import *
from .clock import *
from .clock_io import ClockIO
from .conversions import *
from .interface import *
from .ref import LazyDefnRef, LazyInstRef

# circuit
from .circuit import *

# higher-order operators
from .braid import *

# verilog
from .frontend.verilog import (declare_from_verilog, declare_from_verilog_file,
                               define_from_verilog, define_from_verilog_file,
                               DeclareFromVerilog, DeclareFromVerilogFile,
                               DefineFromVerilog, DefineFromVerilogFile,
                               set_verilog_importer)
from .backend.verilog import *

# compile
from .compile import *

from .logging import root_logger


mantle_target = None
def set_mantle_target(t):
     global mantle_target
     if mantle_target is not None and mantle_target != t:
         root_logger().warning('changing mantle target', mantle_target, t )
     mantle_target = t

from .backend.util import set_codegen_debug_info
from .enum import Enum
import magma.util
import magma.syntax

from .is_primitive import isprimitive
from .is_definition import isdefinition

from .uniquification import UniquificationPass


from hwtypes.bit_vector_abc import TypeFamily

from magma.family import get_family
_Family_ = get_family()

from .generator import Generator, Generator2
from .monitor import MonitorIO, MonitorGenerator, make_monitor_ports
import magma.operators
from .operators import slice
from .inline_verilog import inline_verilog
from .display import display, posedge, negedge, File, time
from .log import info, debug, warning, error



import string
from ast_tools.stack import _SKIP_FRAME_DEBUG_STMT, get_symbol_table
from magma.circuit import _definition_context_stack
from magma.passes.passes import CircuitPass
from magma.verilog_utils import convert_values_to_verilog_str


def _inline_verilog(cls, inline_str, **kwargs):
    format_args = {}
    for key, arg in kwargs.items():
        format_args[key] = convert_values_to_verilog_str(cls, arg)
    cls.inline_verilog_strs.append(inline_str.format(**format_args))


def _process_inline_verilog(cls, format_str, format_args, symbol_table):

    if symbol_table is None:
        _inline_verilog(cls, format_str, **format_args)
        return

    fieldnames = [fname
                  for _, fname, _, _ in string.Formatter().parse(format_str)
                  if fname]
    for field in fieldnames:
        if field in format_args:
            continue
        try:
            value = eval(field, {}, symbol_table)
        except NameError:
            continue
        # These have special handling, don't convert to string.
        value = convert_values_to_verilog_str(cls, value)
        value = value.replace("{", "{{").replace("}", "}}")
        format_str = format_str.replace(f"{{{field}}}", value)
    _inline_verilog(cls, format_str, **format_args)


class ProcessInlineVerilogPass(CircuitPass):
    def __call__(self, cls):
        if cls.inline_verilog_generated:
            return
        cls.inline_verilog_wire_counter = 0
        with cls.open():
            for fields in cls._context_._inline_verilog:
                _process_inline_verilog(cls, *fields)
        if getattr(cls, "instances", []):
            cls._is_definition = True
        cls.inline_verilog_generated = True


def inline_verilog(format_str, **kwargs):
    exec(_SKIP_FRAME_DEBUG_STMT)
    symbol_table = get_symbol_table([inline_verilog], copy_locals=True)

    context = _definition_context_stack.peek()
    context.add_inline_verilog(format_str, kwargs, symbol_table)



from coreir import Context
from ..array import Array
from ..tuple import Tuple
from ..clock import AsyncReset, AsyncResetN, Clock
from ..conversions import convertbit
from ..config import config, EnvConfig
from ..logging import root_logger
from ..t import In, Out
from .. import singleton
from ..circuit import Circuit
from ..interface import IO
from .coreir_transformer import DefnOrDeclTransformer
from ..passes import DefinitionPass
from .util import keydefaultdict
from ..wire import wire


config._register(
    coreir_backend_log_level=EnvConfig(
        "MAGMA_COREIR_BACKEND_LOG_LEVEL", "WARN"),
)

_logger = root_logger().getChild("coreir_backend")
_logger.setLevel(config.coreir_backend_log_level)


# Singleton context meant to be used with coreir/magma code
@singleton
class CoreIRContextSingleton:
    __instance = None

    def get_instance(self):
        return self.__instance

    def reset_instance(self):
        self.__instance = Context()

    def __init__(self):
        self.__instance = Context()


CoreIRContextSingleton()

_context_to_modules = {}


class CoreIRBackend:
    def __init__(self, context=None):
        singleton = CoreIRContextSingleton().get_instance()
        if context is None:
            context = singleton
        if context is not singleton:
            _logger.warning("Creating CoreIRBackend with non-singleton CoreIR "
                            "context.")
        self.modules = _context_to_modules.setdefault(context, {})
        self.context = context
        self.libs = keydefaultdict(self.context.get_lib)
        self.libs_used = set()
        self.constant_cache = {}
        self.sv_bind_files = {}

    def compile(self, defn_or_decl):
        _logger.debug(f"Compiling: {defn_or_decl.name}")
        transformer = DefnOrDeclTransformer(self, defn_or_decl)
        transformer.run()
        self.modules[defn_or_decl.name] = transformer.coreir_module
        return self.modules


class InsertWrapCasts(DefinitionPass):
    def define_wrap(self, wrap_type, in_type, out_type):
        class Wrap(Circuit):
            name = f"coreir_wrap{wrap_type}".replace("(", "").replace(")", "")
            io = IO(**{"in": In(in_type), "out": Out(out_type)})
            coreir_genargs = {"type": wrap_type}
            coreir_name = "wrap"
            coreir_lib = "coreir"

            def simulate(self, value_store, state_store):
                input_val = value_store.get_value(getattr(self, "in"))
                value_store.set_value(self.out, input_val)
        return Wrap

    def wrap_if_named_type(self, port, definition):
        if isinstance(port, (Array, Tuple)):
            for t in port:
                self.wrap_if_named_type(t, definition)
            return
        if not port.is_input():
            return
        if not (isinstance(port, (AsyncReset, AsyncResetN, Clock)) or
                isinstance(port.trace(), (AsyncReset, AsyncResetN, Clock))):
            return
        value = port.trace()
        if value is None or issubclass(type(value), type(port).flip()):
            return
        if isinstance(port, (AsyncReset, AsyncResetN, Clock)):
            T = type(port).flip()
        else:
            T = type(value).flip()

        with definition.open():
            port.unwire(value)
            inst = self.define_wrap(T, type(port), type(value))()
            wire(getattr(inst, "in"), convertbit(value, type(port)))
            wire(inst.out, convertbit(port, type(value)))

    def __call__(self, definition):
        # copy, because wrapping might add instances
        instances = definition.instances[:]
        for instance in definition.instances:
            if type(instance).coreir_name == "wrap" or \
                    type(instance).coreir_name == "unwrap":
                continue
            for port in instance.interface.ports.values():
                self.wrap_if_named_type(port, definition)
        for port in definition.interface.ports.values():
            self.wrap_if_named_type(port, definition)


def compile(main, file_name=None, context=None):
    InsertWrapCasts(main).run()
    backend = CoreIRBackend(context)
    backend.compile(main)
    if file_name is not None:
        return backend.modules[main.coreir_name].save_to_file(file_name)
    else:
        return backend.modules[main.coreir_name]



class PortView:
    def __init__(self, port, parent):
        self.port = port
        self.parent = parent
        self.name = self

    def __getitem__(self, key):
        port = self.port
        try:
            item = port[key]
        except KeyError:
            raise Exception(f"Can only use getitem with arrays and "
                            f"tuples not {type(self.port)}")
        return PortView(item, self.parent)

    def __getattr__(self, key):
        port = self.port
        try:
            attr = getattr(port, key)
        except AttributeError:
            return object.__getattribute__(self, key)
        return PortView(attr, self.parent)


class InstView:
    def __init__(self, inst, parent=None):
        self.inst = inst
        self.circuit = type(inst)
        self.instance_map = None
        if hasattr(self.circuit, "instances"):
            self.instance_map = {instance.name: instance for instance in
                                 self.circuit.instances}
        self.parent = parent

    def __getattr__(self, attr):
        try:
            if attr in self.inst.interface.ports.keys():
                return PortView(self.inst.interface.ports[attr], self)
            elif attr in self.instance_map:
                return InstView(self.instance_map[attr], self)
        except AttributeError:
            return object.__getattribute__(self, attr)



"""
Defines a subtype of m.Array called m.Bits

m.Bits[N] is roughly equivalent ot m.Array[N, T]
"""
import operator
from functools import lru_cache, wraps
import typing as tp
import hwtypes as ht
from hwtypes import BitVector
from hwtypes import AbstractBitVector, AbstractBitVectorMeta, AbstractBit, \
    InconsistentSizeError
from .compatibility import IntegerTypes
from .ref import AnonRef
from .bit import Bit
from .array import Array, ArrayMeta
from .debug import debug_wire
from .t import Type, Direction, In, Out
from magma.circuit import Circuit, coreir_port_mapping, DeclareCoreirCircuit
from magma.family import get_family
from magma.interface import IO
from magma.language_utils import primitive_to_python
from magma.logging import root_logger


_logger = root_logger()


def _coerce(T: tp.Type['Bits'], val: tp.Any) -> 'Bits':
    if not isinstance(val, Bits):
        return T(val)
    elif len(val) != len(T):
        raise InconsistentSizeError('Inconsistent size')
    else:
        return val


def bits_cast(fn: tp.Callable[['Bits', 'Bits'], tp.Any]) -> \
        tp.Callable[['Bits', tp.Any], tp.Any]:
    @wraps(fn)
    def wrapped(self: 'Bits', other: tp.Any) -> tp.Any:
        other = _coerce(type(self), other)
        return fn(self, other)
    return wrapped


class BitsMeta(AbstractBitVectorMeta, ArrayMeta):
    def __new__(mcs, name, bases, namespace, info=(None, None, None), **kwargs):
        return ArrayMeta.__new__(mcs, name, bases, namespace, info, **kwargs)

    def __call__(cls, *args, **kwargs):
        result = super().__call__(*args, **kwargs)
        if all(x.is_output() for x in result.ts) and not cls.is_output():
            if cls.is_input() or cls.is_inout():
                raise TypeError("Can't construct output with input/inout")
            # Make it an output
            return cls.qualify(Direction.Out)(*args, **kwargs)
        return result

    def __getitem__(cls, index):
        if isinstance(index, int):
            index = (index, Bit)
        return ArrayMeta.__getitem__(cls, index)

    def __repr__(cls):
        return str(cls)

    def __str__(cls):
        name = getattr(cls, "orig_name", cls.__name__)
        if not cls.is_concrete:
            return name
        # Handle qualified, unsized e.g. In(Bits)
        if isinstance(cls.T, Direction):
            assert cls.N is None
            return f"{cls.T.name}({name})"

        name += f"[{cls.N}]"
        if cls.is_input():
            name = f"In({name})"
        elif cls.is_output():
            name = f"Out({name})"
        elif cls.is_output():
            name = f"InOut({name})"
        return name


class Bits(Array, AbstractBitVector, metaclass=BitsMeta):
    __hash__ = Array.__hash__
    hwtypes_T = ht.BitVector

    def __init__(self, *args, **kwargs):
        if args and len(args) == 1 and isinstance(args[0], Array) and \
                len(self) > 1 and len(args[0]) <= len(self):
            self.ts = args[0].ts[:]
            # zext for promoting width
            for i in range(len(self) - len(args[0])):
                self.ts.append(Bit(0))
            Type.__init__(self, **kwargs)
        else:
            Array.__init__(self, *args, **kwargs)

    def __repr__(self):
        if not self.name.anon():
            return super().__repr__()
        ts = [repr(t) for t in self.ts]
        return 'bits([{}])'.format(', '.join(ts))

    def bits(self):
        if not self.const():
            raise Exception("Not a constant")

        def _convert(x):
            if x is type(x).VCC:
                return True
            assert x is type(x).GND
            return False

        return [_convert(x) for x in self.ts]

    def __int__(self):
        if not self.const():
            raise Exception("Can't call __int__ on a non-constant")
        return BitVector[len(self)](self.bits()).as_int()

    @debug_wire
    def wire(self, other, debug_info):
        if isinstance(other, IntegerTypes):
            if other.bit_length() > len(self):
                raise ValueError(
                    f"Cannot convert integer {other} "
                    f"(bit_length={other.bit_length()}) to Bits ({len(self)})")
            from .conversions import bits
            other = bits(other, len(self))
        super().wire(other, debug_info)

    @classmethod
    def make_constant(cls, value, num_bits: tp.Optional[int] = None) -> \
            'AbstractBitVector':
        if num_bits is None:
            return cls(value)
        return cls.unsized_t[num_bits](value)

    @classmethod
    @lru_cache(maxsize=None)
    def declare_unary_op(cls, op):
        N = len(cls)
        python_op_name = primitive_to_python(op)

        class _MagmBitsOp(Circuit):
            name = f"magma_Bits_{N}_{op}"
            coreir_name = op
            coreir_lib = "coreir"
            coreir_genargs = {"width": N}
            renamed_ports = coreir_port_mapping
            primitive = True
            stateful = False

            io = IO(I=In(cls), O=Out(cls))

            def simulate(self, value_store, state_store):
                I = cls.hwtypes_T[N](value_store.get_value(self.I))
                O = int(getattr(operator, python_op_name)(I))
                value_store.set_value(self.O, O)

        return _MagmBitsOp

    @classmethod
    @lru_cache(maxsize=None)
    def declare_binary_op(cls, op):
        N = len(cls)
        python_op_name = primitive_to_python(op)
        python_op = getattr(operator, python_op_name)

        class _MagmBitsOp(Circuit):
            name = f"magma_Bits_{N}_{op}"
            coreir_name = op
            coreir_lib = "coreir"
            coreir_genargs = {"width": N}
            renamed_ports = coreir_port_mapping
            primitive = True
            stateful = False

            io = IO(I0=In(cls), I1=In(cls), O=Out(cls))

            def simulate(self, value_store, state_store):
                I0 = cls.hwtypes_T[N](value_store.get_value(self.I0))
                I1 = cls.hwtypes_T[N](value_store.get_value(self.I1))
                O = int(python_op(I0, I1))
                value_store.set_value(self.O, O)

        return _MagmBitsOp

    @classmethod
    @lru_cache(maxsize=None)
    def declare_compare_op(cls, op):
        N = len(cls)
        python_op_name = primitive_to_python(op)

        class _MagmBitsOp(Circuit):
            name = f"magma_Bits_{N}_{op}"
            coreir_name = op
            coreir_lib = "coreir"
            coreir_genargs = {"width": N}
            renamed_ports = coreir_port_mapping
            primitive = True
            stateful = False

            io = IO(I0=In(cls), I1=In(cls), O=Out(Bit))

            def simulate(self, value_store, state_store):
                I0 = cls.hwtypes_T[N](value_store.get_value(self.I0))
                I1 = cls.hwtypes_T[N](value_store.get_value(self.I1))
                O = int(getattr(operator, python_op_name)(I0, I1))
                value_store.set_value(self.O, O)

        return _MagmBitsOp

    @classmethod
    @lru_cache(maxsize=None)
    def declare_ite(cls, T):
        t_str = str(T)
        # Sanitize
        t_str = t_str.replace("(", "_")
        t_str = t_str.replace(")", "")
        t_str = t_str.replace("[", "_")
        t_str = t_str.replace("]", "")
        N = len(cls)

        class _MagmBitsOp(Circuit):
            name = f"magma_Bits_{N}_ite_{t_str}"
            coreir_name = "mux"
            coreir_lib = "coreir"
            coreir_genargs = {"width": len(T)}
            renamed_ports = coreir_port_mapping
            primitive = True
            stateful = False

            io = IO(I0=In(T), I1=In(T), S=In(Bit), O=Out(T))

            def simulate(self, value_store, state_store):
                I0 = cls.hwtypes_T[N](value_store.get_value(self.I0))
                I1 = cls.hwtypes_T[N](value_store.get_value(self.I1))
                S = ht.Bit(value_store.get_value(self.S))
                O = I1 if S else I0
                value_store.set_value(self.O, O)

        return _MagmBitsOp

    def bvnot(self) -> 'AbstractBitVector':
        return self.declare_unary_op("not")()(self)

    @bits_cast
    def bvand(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("and")()(self, other)

    @bits_cast
    def bvor(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("or")()(self, other)

    @bits_cast
    def bvxor(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("xor")()(self, other)

    @bits_cast
    def bvshl(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("shl")()(self, other)

    @bits_cast
    def bvlshr(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("lshr")()(self, other)

    def bvashr(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def bvrol(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def bvror(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def bvcomp(self, other) -> 'AbstractBitVector[1]':
        return Bits[1](self == other)

    def bveq(self, other) -> AbstractBit:
        return self.declare_compare_op("eq")()(self, other)

    def bvult(self, other) -> AbstractBit:
        raise NotImplementedError()

    def bvule(self, other) -> AbstractBit:
        # For wiring
        if not self.is_output():
            return Type.__le__(self, other)
        raise NotImplementedError()

    def bvugt(self, other) -> AbstractBit:
        raise NotImplementedError()

    def bvuge(self, other) -> AbstractBit:
        raise NotImplementedError()

    def bvslt(self, other) -> AbstractBit:
        raise NotImplementedError()

    def bvsle(self, other) -> AbstractBit:
        raise NotImplementedError()

    def bvsgt(self, other) -> AbstractBit:
        raise NotImplementedError()

    def bvsge(self, other) -> AbstractBit:
        raise NotImplementedError()

    def bvneg(self) -> 'AbstractBitVector':
        raise NotImplementedError()

    def adc(self, other, carry) -> tp.Tuple['AbstractBitVector', AbstractBit]:
        raise NotImplementedError()

    def ite(self, t_branch, f_branch) -> 'AbstractBitVector':
        type_ = type(t_branch)
        if type_ != type(f_branch):
            raise TypeError("ite expects same type for both branches")
        return self.declare_ite(type_)()(t_branch, f_branch,
                                         self != self.make_constant(0))

    def bvadd(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def bvsub(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def bvmul(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def bvudiv(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def bvurem(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def bvsdiv(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def bvsrem(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def repeat(self, other) -> 'AbstractBitVector':
        r = int(other)
        if r <= 0:
            raise ValueError()

        return type(self).unsized_t[r * self.size](r * self.ts)

    def sext(self, other) -> 'AbstractBitVector':
        raise NotImplementedError()

    def ext(self, other) -> 'AbstractBitVector':
        return self.zext(other)

    def zext(self, other) -> 'AbstractBitVector':
        ext = int(other)
        if ext < 0:
            raise ValueError()

        T = type(self).unsized_t
        return self.concat(T[ext](0))

    def __invert__(self):
        return self.bvnot()

    def __and__(self, other):
        try:
            return self.bvand(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __or__(self, other):
        try:
            return self.bvor(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __xor__(self, other):
        try:
            return self.bvxor(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __lshift__(self, other):
        try:
            return self.bvshl(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __rshift__(self, other):
        try:
            return self.bvlshr(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __neg__(self):
        return self.bvneg()

    def __add__(self, other):
        try:
            return self.bvadd(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __sub__(self, other):
        try:
            return self.bvsub(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __mul__(self, other):
        try:
            return self.bvmul(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __floordiv__(self, other):
        try:
            return self.bvudiv(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __truediv__(self, other):
        try:
            return self.bvudiv(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __mod__(self, other):
        try:
            return self.bvurem(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __eq__(self, other):
        try:
            return self.bveq(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __ne__(self, other):
        try:
            return self.bvne(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __ge__(self, other):
        try:
            return self.bvuge(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __gt__(self, other):
        try:
            return self.bvugt(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __le__(self, other):
        try:
            return self.bvule(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __lt__(self, other):
        try:
            return self.bvult(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    @classmethod
    def get_family(cls):
        return get_family()

    @classmethod
    def unflatten(cls, value):
        size = len(value)
        if cls.N != size:
            _logger.warning(f"Using {repr(cls)} to unflatten {size} elements")
        return Bits[size](value)

    def unused(self):
        if self.is_input() or self.is_inout():
            raise TypeError("unused cannot be used with input/inout")
        if not getattr(self, "_magma_unused_", False):
            DefineUnused(len(self))().I.wire(self)
            # "Cache" unused calls so only one is produced
            self._magma_unused_ = True

    def undriven(self):
        if self.is_output() or self.is_inout():
            raise TypeError("undriven cannot be used with output/inout")
        self.wire(DefineUndriven(len(self))().O)

    def __getitem__(self, index):
        if isinstance(index, Bits):
            if len(index) > len(self):
                raise TypeError(f"Unexpected index width: {len(index)}")
            if len(index) < len(self):
                index = index.zext(len(self) - len(index))
            return (self >> index)[0]
        return super().__getitem__(index)


def make_Define(name, port, direction):
    def simulate(self, value_store, state_store):
        pass

    @lru_cache(maxsize=None)
    def Define(width):
        return DeclareCoreirCircuit(
            name,
            port, direction(Bits[width]),
            coreir_name=name,
            coreir_lib="coreir",
            coreir_genargs={"width": width},
            simulate=simulate
        )
    return Define


DefineUndriven = make_Define("undriven", "O", Out)
DefineUnused = make_Define("term", "I", In)


BitsType = Bits


class UInt(Bits):
    hwtypes_T = ht.UIntVector

    @bits_cast
    def bvult(self, other) -> AbstractBit:
        return self.declare_compare_op("ult")()(self, other)

    @bits_cast
    def bvule(self, other) -> AbstractBit:
        # For wiring
        if self.is_input():
            return Type.__le__(self, other)
        return self.declare_compare_op("ule")()(self, other)

    @bits_cast
    def bvugt(self, other) -> AbstractBit:
        return self.declare_compare_op("ugt")()(self, other)

    @bits_cast
    def bvuge(self, other) -> AbstractBit:
        return self.declare_compare_op("uge")()(self, other)

    @bits_cast
    def bvadd(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("add")()(self, other)

    @bits_cast
    def bvsub(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("sub")()(self, other)

    @bits_cast
    def bvmul(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("mul")()(self, other)

    @bits_cast
    def bvudiv(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("udiv")()(self, other)

    @bits_cast
    def bvurem(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("urem")()(self, other)

    def adc(self, other: 'Bits', carry: Bit) -> tp.Tuple['Bits', Bit]:
        """
        add with carry
        returns a two element tuple of the form (result, carry)
        """
        T = type(self)
        other = _coerce(T, other)
        carry = _coerce(T.unsized_t[1], carry)

        a = self.zext(1)
        b = other.zext(1)
        c = carry.zext(T.size)

        res = a + b + c
        return res[0:-1], res[-1]


class SInt(Bits):
    hwtypes_T = ht.SIntVector

    def __init__(self, *args, **kwargs):
        if args and len(args) == 1 and isinstance(args[0], Array) and \
                len(self) > 1 and len(args[0]) <= len(self):
            self.ts = args[0].ts[:]
            # zext for promoting width
            for i in range(len(self) - len(args[0])):
                self.ts.append(args[0].ts[-1])
            Type.__init__(self, **kwargs)
        else:
            Array.__init__(self, *args, **kwargs)

    @bits_cast
    def bvslt(self, other) -> AbstractBit:
        return self.declare_compare_op("slt")()(self, other)

    @bits_cast
    def bvsle(self, other) -> AbstractBit:
        # For wiring
        if self.is_input():
            return Type.__le__(self, other)
        return self.declare_compare_op("sle")()(self, other)

    @bits_cast
    def bvsgt(self, other) -> AbstractBit:
        return self.declare_compare_op("sgt")()(self, other)

    @bits_cast
    def bvsge(self, other) -> AbstractBit:
        return self.declare_compare_op("sge")()(self, other)

    @bits_cast
    def bvadd(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("add")()(self, other)

    @bits_cast
    def bvsub(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("sub")()(self, other)

    @bits_cast
    def bvmul(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("mul")()(self, other)

    @bits_cast
    def bvsdiv(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("sdiv")()(self, other)

    @bits_cast
    def bvsrem(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("srem")()(self, other)

    def bvneg(self) -> 'AbstractBitVector':
        return self.declare_unary_op("neg")()(self)

    @bits_cast
    def bvashr(self, other) -> 'AbstractBitVector':
        return self.declare_binary_op("ashr")()(self, other)

    def __mod__(self, other):
        try:
            return self.bvsrem(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __floordiv__(self, other):
        try:
            return self.bvsdiv(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __truediv__(self, other):
        try:
            return self.bvsdiv(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __ge__(self, other):
        try:
            return self.bvsge(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __gt__(self, other):
        try:
            return self.bvsgt(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __le__(self, other):
        try:
            return self.bvsle(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __lt__(self, other):
        try:
            return self.bvslt(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def __rshift__(self, other):
        try:
            return self.bvashr(other)
        except InconsistentSizeError as e:
            raise e from None
        except TypeError:
            return NotImplemented

    def adc(self, other: 'Bits', carry: Bit) -> tp.Tuple['Bits', Bit]:
        """
        add with carry
        returns a two element tuple of the form (result, carry)
        """
        T = type(self)
        other = _coerce(T, other)
        carry = _coerce(T.unsized_t[1], carry)

        a = self.sext(1)
        b = other.sext(1)
        c = carry.zext(T.size)

        res = a + b + c
        return res[0:-1], res[-1]

    def sext(self, other) -> 'AbstractBitVector':
        ext = int(other)
        if ext < 0:
            raise ValueError()

        T = type(self).unsized_t
        return self.concat(T[ext]([self[-1] for _ in range(ext)]))



import os
from tempfile import NamedTemporaryFile

from .simulator import CircuitSimulator, ExecutionState
from ..backend import coreir_, coreir_utils
from ..frontend.coreir_ import GetMagmaContext
from ..scope import Scope
from ..ref import DefnRef, ArrayRef, TupleRef
from ..array import Array
from ..tuple import Tuple
from ..bit import Bit
from ..bitutils import int2seq
from ..clock import Clock
from ..transforms import flatten
from ..circuit import CircuitType
from ..uniquification import uniquification_pass, UniquificationMode
from ..passes.clock import WireClockPass

import coreir

__all__ = ['CoreIRSimulator']

def is_defn_bit(bit):
    name = bit.name
    if isinstance(name, ArrayRef):
        name = name.array.name

    return isinstance(name, DefnRef)

def convert_to_coreir_path(bit, scope):
    insts = []

    # Handle bits attached to nested definitions
    if scope.inst is not None and is_defn_bit(bit):
        defn_bit = bit
        cur_inst = scope.inst
        scope = scope.parent
        if isinstance(defn_bit.name, ArrayRef):
            bit = cur_inst.interface[defn_bit.name.array.name.name]
            bit = bit[defn_bit.name.index]
        else:
            bit = cur_inst.interface[defn_bit.name.name]

    # Build instance path to this bit
    while scope.inst is not None:
        assert(isinstance(scope.inst, CircuitType))
        inst_name = scope.inst.name
        insts.insert(0, inst_name)
        scope = scope.parent

    last_component = coreir_utils.magma_port_to_coreir_port(bit)
    last_inst, port = last_component.split('.', 1)
    insts.append(last_inst)

    # Handle renaming due to flatten types
    def flattened_name(name):
        if isinstance(name, DefnRef):
            return str(name)
        if isinstance(name, ArrayRef):
            array_name = flattened_name(name.array.name)
            # CoreIR simulator doesn't flatten array of bits
            if issubclass(name.array.T, Bit):
                return f"{array_name}.{name.index}"
            else:
                return f"{array_name}_{name.index}"
        if isinstance(name, TupleRef):
            tuple_name = flattened_name(name.tuple.name)
            index = name.index
            try:
                int(index)
                # If it's an int, insert `_` prefix because coreir doesn't
                # allow fields to start with an int
                index = f"_{index}"
            except ValueError:
                pass
            return f"{tuple_name}_{index}"
        raise NotImplementedError(name, type(name))
    port = flattened_name(bit.name)

    ports = [port]

    return insts, ports

# This exists because things like clocks and setting values doesn't have the new API
# so convert from what the new API expects to what the old API expects
def old_style_path(insts, ports):
    last_inst = insts[-1]
    return insts[:-1] + [last_inst + '.' + ports[0]]

class WatchPoint:
    idx = 0
    def __init__(self, bit, scope, simulator, value):
        self.bit = bit
        self.scope = scope
        self.simulator = simulator
        self.value = value
        self.old_val = self.simulator.get_value(self.bit, self.scope)

        WatchPoint.idx += 1
        self.idx = WatchPoint.idx

    def was_triggered(self):
        new_val = self.simulator.get_value(self.bit, self.scope)
        triggered = new_val != self.old_val
        if self.value:
            if self.value != new_val:
                triggered = False
        self.old_val = new_val

        return triggered

class CoreIRSimulator(CircuitSimulator):
    def __get_cur_cycles(self):
        instpath, port_selects = convert_to_coreir_path(self.clock, Scope())
        steps = self.simulator_state.get_clock_cycles(old_style_path(instpath, port_selects))
        return steps

    def __get_clock_value(self):
        if self.clock is None:
            return None
        return self.get_value(self.clock, Scope())

    def __get_triggered_points(self):
        triggered = []
        for watch in self.watchpoints:
            if watch.was_triggered():
                triggered.append(watch)

        return triggered

    def __init__(self, circuit, clock, coreir_filename=None, context=None,
                 namespaces=["global"], opts={}):
        uniquification_mode_str = opts.get("uniquify", "UNIQUIFY")
        uniquification_mode = getattr(UniquificationMode,
                                      uniquification_mode_str, None)
        if uniquification_mode is None:
            raise ValueError(f"Invalid uniquification mode "
                             f"{uniquification_mode_str}")
        uniquification_pass(circuit, uniquification_mode)
        self.watchpoints = []
        self.default_scope = Scope()

        need_cleanup = False
        if not coreir_filename:
            coreir_file = NamedTemporaryFile(delete=False, suffix='.json')
            coreir_file.close()
            coreir_filename = coreir_file.name
            need_cleanup = True

        self.clock = clock
        WireClockPass(circuit).run()

        if context is None:
            self.ctx = GetMagmaContext()
        else:
            self.ctx = context
        coreir_.compile(circuit, coreir_filename, context=self.ctx)

        # Initialize interpreter, get handle back to interpreter state
        self.ctx.get_lib("commonlib")
        self.ctx.enable_symbol_table()
        coreir_circuit = self.ctx.load_from_file(coreir_filename)
        self.ctx.run_passes(["rungenerators", "wireclocks-coreir", "verifyconnectivity --noclkrst",
                             "flattentypes", "flatten", "verifyconnectivity --noclkrst", "deletedeadinstances"],
                            namespaces=namespaces)
        self.simulator_state = coreir.SimulatorState.make(coreir_circuit)

        if need_cleanup:
            os.remove(coreir_filename)

        def create_zeros_init(arrOrTuple):
            if isinstance(arrOrTuple, Array):
                return [create_zeros_init(el) for el in arrOrTuple.ts]
            elif isinstance(arrOrTuple, Tuple):
                return {k: create_zeros_init(v) for k,v in zip(arrOrTuple.keys(), arrOrTuple.values())}
            else:
                return 0

        # Need to set values for all circuit inputs or interpreter crashes
        for topin in circuit.interface.outputs():
            if not isinstance(topin, Clock):
                arr = topin
                init = create_zeros_init(arr)

                self.set_value(topin, init, Scope())

        if clock is not None:
            insts, ports = convert_to_coreir_path(clock, Scope())
            self.simulator_state.set_main_clock(old_style_path(insts, ports))

        self.simulator_state.reset_circuit()

        for topin in circuit.interface.outputs():
            if isinstance(topin, Clock):
                insts, ports = convert_to_coreir_path(topin, Scope())
                self.simulator_state.set_clock_value(old_style_path(insts, ports), True, False)

        self.evaluate()

    def get_capabilities(self):
        # TBD
        pass

    def get_value(self, bit, scope=None):
        if scope is None:
            scope = self.default_scope
        if bit.const():
            return True if bit == VCC else False

        # Symbol table doesn't support arrays of arrays
        if isinstance(bit, Array) and (isinstance(bit[0], Array) or isinstance(bit[0], Tuple)):
            r = []
            for arr in bit:
                r.append(self.get_value(arr, scope))
            return r
        elif isinstance(bit, Tuple):
            r = {}
            for k,v in zip(bit.keys(), bit.values()):
                r[k] = self.get_value(v, scope)
            return r
        else:
            insts, ports = convert_to_coreir_path(bit, scope)

            bools = self.simulator_state.get_value(insts, ports)
            if len(bools) == 1:
                return bools[0]
            return bools

    def set_value(self, bit, newval, scope=None):
        if scope is None:
            scope = self.default_scope
        if isinstance(bit, Array) and len(newval) != len(bit):
                raise ValueError(f"Excepted a value of lengh {len(bit)} not"
                                 f" {len(newval)}")
        if isinstance(bit, Array) and (isinstance(bit[0], Array) or isinstance(bit[0], Tuple)):
            for i, arr in enumerate(bit):
                self.set_value(arr, newval[i], scope)
        elif isinstance(bit, Tuple):
            for k,v in zip(bit.keys(), bit.values()):
                self.set_value(v, newval[k], scope)
        else:
            insts, ports = convert_to_coreir_path(bit, scope)
            self.simulator_state.set_value(old_style_path(insts, ports), newval)

    def evaluate(self, no_update=False):
        clkvalue = self.__get_clock_value()
        if clkvalue is not None:
            insts, ports = convert_to_coreir_path(self.clock, Scope())
            self.simulator_state.set_clock_value(old_style_path(insts, ports), False, False)
        self.simulator_state.execute()
        if clkvalue is not None:
            self.simulator_state.set_clock_value(old_style_path(insts, ports), not clkvalue, clkvalue)

        return ExecutionState(triggered_points=self.__get_triggered_points(), clock=clkvalue, cycles=0)

    def advance(self, halfcycles=1):
        cycles = self.__get_cur_cycles()
        # TODO add a function to interpreter to avoid doing this for loop in python
        watchpoints = []
        for i in range(0, halfcycles):
            self.simulator_state.run_half_cycle()
            watchpoints = self.__get_triggered_points()
            if len(watchpoints) > 0:
                break

        post_cycles = self.__get_cur_cycles()
        return ExecutionState(triggered_points=watchpoints, clock=self.__get_clock_value(), cycles=post_cycles - cycles)

    def rewind(self, halfcycles):
        self.simulator_state.rewind(halfcycles)
        return ExecutionState(triggered_points=self.__get_triggered_points(), clock=self.__get_clock_value(), cycles=0)

    def cont(self):
        pre_cycles = self.__get_cur_cycles()
        self.simulator_state.run()
        post_cycles = self.__get_cur_cycles()
        return ExecutionState(triggered_points=self.__get_triggered_points(), clock=self.__get_clock_value(), cycles=(post_cycles - pre_cycles))

    def add_watchpoint(self, bit, scope, value=None):
        if value is None:
            raise Exception("CoreIR Simulator does not support watching for value change")

        insts, ports = convert_to_coreir_path(bit, scope)
        self.simulator_state.set_watchpoint(insts, ports, value)

        self.watchpoints.append(WatchPoint(bit, scope, self, value))

        return self.watchpoints[-1].idx

    def delete_watchpoint(self, num):
        for i, w in enumerate(self.watchpoints):
            if w.idx == num:
                insts, ports = convert_to_coreir_path(w.bit, w.scope)
                self.simulator_state.delete_watchpoint(insts, ports)
                del self.watchpoints[i]
                return True

        return False



from .passes import EditCircuitPass
from ..is_definition import isdefinition


def _drive_undriven(interface):
    undriven = False
    for port in interface.ports.values():
        if port.is_input() and port.trace() is None:
            undriven = True
            port.undriven()
    return undriven


class DriveUndrivenPass(EditCircuitPass):
    def edit(self, circuit):
        if _drive_undriven(circuit.interface):
            circuit._is_definition = True
        if not isdefinition(circuit):
            return
        for inst in circuit.instances:
            _drive_undriven(inst.interface)



from magma.array import Array
from magma.circuit import Circuit
from magma.digital import Digital
from magma.ref import DefnRef, InstRef, NamedRef, ArrayRef, TupleRef
from magma.t import Type, Direction
from magma.tuple import Tuple
from magma.view import InstView, PortView


def value_to_verilog_name(value):
    if isinstance(value, Array) and not issubclass(value.T, Digital):
        elems = ", ".join(value_to_verilog_name(t) for t in reversed(value))
        return f"'{{{elems}}}"
    elif isinstance(value, Tuple):
        raise NotImplementedError("Inlining unflattened tuple")
    elif isinstance(value, PortView):
        parent_name = value_to_verilog_name(value.parent)
        return f"{parent_name}.{verilog_name(value.port.name)}"
    elif isinstance(value, InstView):
        prefix = ""
        if isinstance(value.parent, Circuit):
            prefix = value.parent.name + "."
        else:
            prefix = value_to_verilog_name(value.parent) + "."
        return prefix + value.inst.name
    return verilog_name(value.name)


def verilog_name(name, inst_sep="_"):
    if isinstance(name, DefnRef):
        return str(name)
    if isinstance(name, InstRef):
        return f"{name.inst.name}{inst_sep}{str(name)}"
    if isinstance(name, NamedRef):
        return str(name)
    if isinstance(name, ArrayRef):
        array_name = verilog_name(name.array.name)
        if issubclass(name.array.T, Digital):
            index = f"[{name.index}]"
        else:
            index = f"_{name.index}"
        return f"{array_name}{index}"
    if isinstance(name, TupleRef):
        tuple_name = verilog_name(name.tuple.name)
        index = name.index
        try:
            int(index)
            # python/coreir don't allow pure integer names
            index = f"_{index}"
        except ValueError:
            pass
        return f"{tuple_name}_{index}"
    raise NotImplementedError(name, type(name))


def _get_top_level_ref(ref):
    if isinstance(ref, ArrayRef):
        return _get_top_level_ref(ref.array.name)
    if isinstance(ref, TupleRef):
        return _get_top_level_ref(ref.tuple.name)
    return ref


def _sanitize(name):
    return name.replace(".", "_").replace("[", "_").replace("]", "")


def convert_values_to_verilog_str(cls, value):
    if isinstance(value, Type):
        if value.is_input():
            value = value.value()
        if not isinstance(_get_top_level_ref(value.name), DefnRef):
            if not hasattr(value, "_magma_inline_wire_"):
                # Insert a wire so it can't be inlined out
                temp_name = f"_magma_inline_wire"
                temp_name += f"{cls.inline_verilog_wire_counter}"
                with cls.open():
                    temp = type(value).qualify(Direction.Undirected)(
                        name=temp_name
                    )
                    cls.inline_verilog_wire_counter += 1
                    temp @= value
                    temp.unused()
                    value._magma_inline_wire_ = temp
            value = value._magma_inline_wire_
        return value_to_verilog_name(value)
    if isinstance(value, PortView):
        if value.port.is_input():
            raise NotImplementedError()
        if not hasattr(value, "_magma_inline_wire_"):
            ref = _get_top_level_ref(value.port.name)
            if isinstance(ref, InstRef):
                defn = ref.inst.defn
            else:
                assert isinstance(ref, DefnRef)
                defn = ref.defn
            with defn.open():
                temp = type(value.port).qualify(Direction.Undirected)(
                    name=f"_magma_inline_wire{cls.inline_verilog_wire_counter}"
                )
                cls.inline_verilog_wire_counter += 1
                temp @= value.port
                temp.unused()
                temp = PortView(temp, value.parent)
                value._magma_inline_wire_ = temp
            value = value._magma_inline_wire_
        return value_to_verilog_name(value)
    return str(value)
